#include <SD.h>
#include <SPI.h>

Sd2Card card;
SdVolume volume;
SdFile root;
SdFile track;

#define BUFFER_SIZE 32

uint8_t mp3_buffer[BUFFER_SIZE];
boolean need_data = 1;
boolean sd_present = 0;
boolean playing = 1;
int track_num = 1;
int track_count = 0;
char track_name[] = "001.mp3";
unsigned char mp3_sample = 0;
unsigned int mp3_sample_repetition = 0;

#define LED 6

#define CLOCK_OUT  7 // PD7
#define DATA_IN  5 // PD5

// VS1011 chip pins
#define MP3_CS 16 // PC2
#define MP3_DCS 17 // PC3
#define MP3_RESET 18 // PC4
#define MP3_DREQ 19 // PC5

#define SD_CS 10

//VS10xx SCI Registers
#define SCI_MODE 0x00
#define SCI_STATUS 0x01
#define SCI_BASS 0x02
#define SCI_CLOCKF 0x03
#define SCI_DECODE_TIME 0x04
#define SCI_AUDATA 0x05
#define SCI_WRAM 0x06
#define SCI_WRAMADDR 0x07
#define SCI_HDAT0 0x08
#define SCI_HDAT1 0x09
#define SCI_AIADDR 0x0A
#define SCI_VOL 0x0B
#define SCI_AICTRL0 0x0C
#define SCI_AICTRL1 0x0D
#define SCI_AICTRL2 0x0E
#define SCI_AICTRL3 0x0F

void setup()
{
  Serial.begin(9600);
  
  pinMode(LED, OUTPUT);
  
  // Flash the LED to indicate program start
  for (int i = 0; i < 20; ++i)
  {
    digitalWrite(LED, LOW);
    delay(20);
    digitalWrite(LED, HIGH);
    delay(20);
  }

  // Set up the interface to the capacitive sensing board
  pinMode(CLOCK_OUT, OUTPUT);
  digitalWrite(CLOCK_OUT, HIGH);
  pinMode(DATA_IN, INPUT);
  digitalWrite(DATA_IN, HIGH); // Turn on internal pull-up

  // Initialize the pins that talk to the MP3 decoder chip
  pinMode(MP3_CS, OUTPUT);
  pinMode(MP3_DCS, OUTPUT);
  pinMode(MP3_RESET, OUTPUT);
  pinMode(MP3_DREQ, INPUT);
  
  // Tell the Mp3 chip to stop taking over the SPI bus.
  digitalWrite(MP3_CS, HIGH);
  digitalWrite(MP3_DCS, HIGH);
  
  pinMode(SD_CS, OUTPUT); //Pin 10 must be set as an output for the SD communication to work.

  // Bring up the SD card
  if (sd_init()) {
    sd_present = 1;
    Serial.println("SD card found!");
    count_tracks();
    open_track();
  }

  // MP3 decoder is running at 12.8 MHz and has a max SPI speed of 1/6 clk = 2.13 MHz
  SPI.setClockDivider(SPI_CLOCK_DIV8); // SPI speed of 1 MHz to be safe
  SPI.transfer(0xFF); // Dummy byte
  delay(10); // Not sure why this is here; copied from Sparkfun sample code
  digitalWrite(MP3_RESET, HIGH); //Bring up VS1011e
  
  // Activate the clock doubler and tell the VS1011e the input clock speed.
  Mp3WriteRegister(SCI_CLOCKF, 0x9800);
  SPI.setClockDivider(SPI_CLOCK_DIV4);
  Mp3WriteRegister(SCI_VOL, 0x0000); // MAX VOLUME!
  Mp3WriteRegister(SCI_MODE, 1 << 11); // Turn on VS1002 native SPI mode for data transmission

}

boolean sd_init()
{
  return card.init(SPI_FULL_SPEED) && volume.init(&card) && root.openRoot(&volume);
}

boolean open_track()
{
  sprintf(track_name, "%03d.mp3", track_num); //Splice the new file number into this file name
  if (!track.open(&root, track_name, O_READ)) { //Open the file in read mode.
    Serial.print("Failed to open ");
    Serial.println(track_name);
    track.close();
    return 0;
  }
  
  Serial.print("Opened file ");
  Serial.println(track_name);
  
  return 1; // Success!
}

char get_char()
{
  char c = 0;
  for (int i = 0; i < 8; ++i) {
    digitalWrite(CLOCK_OUT, LOW);
    delayMicroseconds(5);
    c = (c << 1) + digitalRead(DATA_IN);
    digitalWrite(CLOCK_OUT, HIGH);
    delayMicroseconds(5);
  }
  return c;
}
  
  
void loop()
{
  // Look for the SD card
  if (!sd_present) {
    if (sd_init()) {
      sd_present = 1;
      ("SD card found!");
      count_tracks();
      open_track();
    }
    else {
      Serial.println("SD card not found!");
      delay(500);
      return;
    }
  }
  
  // If the MP3 decoder wants more data and we have it available
  if (digitalRead(MP3_DREQ) && !need_data && playing) {
    digitalWrite(MP3_DCS, LOW); // Select the data input
    
    // Dump the local data buffer into the VS1011e
    for (int i = 0; i < BUFFER_SIZE; ++i) {
      SPI.transfer(mp3_buffer[i]);
    }
        
    digitalWrite(MP3_DCS, HIGH); // Deselect data input
    

/*    
    // Silly trick to detect if the SD card has been unplugged:
    // If 4096 of the mp3 samples in a row are identical, then we
    // conclude that the card is actually unplugged and is delivering
    // poorly cached data.
    for(int i = 0; i < BUFFER_SIZE; ++i) {
      if (mp3_buffer[i] == mp3_sample)
        mp3_sample_repetition++;
      else {
        mp3_sample = mp3_buffer[0];
        mp3_sample_repetition = 0;
      }
    }
    if (mp3_sample_repetition > 4096) {
      Serial.println("SD card unplugged!");
      Serial.println(mp3_sample_repetition);
      sd_present = 0;
      track.close();
      root.close();
    }
*/

    need_data = 1;
  }
  
  // If the Atmega's data buffer is empty, then load in more data.
  if (need_data && sd_present) {
    // If we've loaded all of the data in the file, then start the next song in the playlist.
    if(!track.read(mp3_buffer, BUFFER_SIZE)) {
      next_track();
    } else {
      need_data = 0;
    }
  }
  
  
  // If the button daughterboard is trying to send us a message, listen to it!
  char c = 0;
  if (digitalRead(DATA_IN) == LOW) {
    c = get_char();
    Serial.print("Command from daughterboard: ");
    Serial.println(c);
  }
  else if (Serial.available() > 0)
    // read the incoming byte:
    c = Serial.read();
    
  if (c == 'n') {
    Serial.println("Next track");
    next_track();
  } else if (c == 's') {
    if (playing)
      Serial.println("Stopping");
    else
      Serial.println("Starting");
    playing = !playing;
  } else if (c == 'b') {
    Serial.println("Previous track");
    prev_track();
  } else if (c == 'p')
    volume_up();
  else if (c == 'm')
    volume_down();
}

void next_track() {
  track.close();

  track_num++;
  if (track_num > track_count)
    track_num = 1;

  open_track();
  need_data = 1;

}

void prev_track() {
  track.close();

  track_num--;
  if (track_num == 0)
    track_num = track_count;

  open_track();
  need_data = 1;
}

void count_tracks() {
  Serial.println("Counting tracks:");
  track_num = 1;

  while (open_track())
  {
    track_num++;
    track.close();
  }

  track_count = track_num - 1;
  track_num = 1; // Go back to the first song on the playlist
  
  Serial.print("Detected ");
  Serial.print(track_count);
  Serial.println(" tracks.");
}

void volume_up() {
  unsigned int volumes = Mp3ReadRegister(SCI_VOL);
  unsigned char volume = volumes & 0xFF;
  Serial.print("Changing volume from -");
  Serial.print(volume);
  if (volume > 0)
    volume -= 6;
  Serial.print(" to -");
  Serial.println(volume);
  volumes = volume + (volume << 8);
  Mp3WriteRegister(SCI_VOL, volumes);
}

void volume_down() {
  unsigned int volumes = Mp3ReadRegister(SCI_VOL);
  unsigned char volume = volumes & 0xFF;
  Serial.print("Changing volume from -");
  Serial.print(volume);
  if (volume < 72)
    volume += 6;
  Serial.print(" to -");
  Serial.println(volume);
  volumes = volume + (volume << 8);
  Mp3WriteRegister(SCI_VOL, volumes);
}

unsigned int Mp3ReadRegister(unsigned char address)
{
  while (!digitalRead(MP3_DREQ))
    ; // Wait for the chip to be ready
  
  digitalWrite(MP3_CS, LOW); // Pull down the chip command select line
  
  SPI.transfer(0x03); // Read command
  SPI.transfer(address);
  
  unsigned int result = SPI.transfer(0xFF); // Dummy transfer to read in data
  result <<= 8; // Shift over to make room for the next byte
  while (!digitalRead(MP3_DREQ)); // Wait for the chip to be ready
  result += SPI.transfer(0xFF); // Another dummy transfer
  while (!digitalRead(MP3_DREQ)); // Wait for the chip to be ready

  digitalWrite(MP3_CS, HIGH); // Deselect the chip

  return result;
}

void Mp3WriteRegister(unsigned char address, unsigned int data)
{
  while (!digitalRead(MP3_DREQ))
    ; // Wait for the chip to be ready
  
  digitalWrite(MP3_CS, LOW); // Pull down the chip command select line
  
  SPI.transfer(0x02); // Write command
  SPI.transfer(address);
  SPI.transfer(data >> 8); // Write the upper 8 bits
  SPI.transfer(data & 255); // Write the lower 8 bits
  
  digitalWrite(MP3_CS, HIGH); // And deselect the chip
}
